/* The copyright in this software is being made available under the BSD
* License, included below. This software may be subject to other third party
* and contributor rights, including patent rights, and no such rights are
* granted under this license.
*
* Copyright (c) 2010-2022, ITU/ISO/IEC
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*  * Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
*  * Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
*  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
*    be used to endorse or promote products derived from this software without
*    specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
* THE POSSIBILITY OF SUCH DAMAGE.
*/

/** \file     TAppMmctsExtCfg.h
\brief    MCTS Extractor configuration class (header)
*/
#ifndef __TAPPMCTSEXTCFG__
#define __TAPPMCTSEXTCFG__

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "TLibCommon/CommonDef.h"
#include <vector>

#if MCTS_EXTRACTION

//! \ingroup TAppMctsExt
//! \{

// ====================================================================================================================
// Class definition
// ====================================================================================================================

/// MCTS Extraction configuration class
class TAppMctsExtCfg
{
protected:
  std::string   m_inputBitstreamFileName;             ///< input bitstream file name
  std::string   m_outputBitstreamFileName;            ///< output subbitstream filename
  Int           m_targetMctsIdx;                      ///< MCTS Id extracted

public:
  TAppMctsExtCfg()
    : m_inputBitstreamFileName()
    , m_outputBitstreamFileName()
    , m_targetMctsIdx(-1)
  {

  }

  virtual ~TAppMctsExtCfg() {}

  Bool  parseCfg(Int argc, TChar* argv[]);   ///< initialize option class from configuration
};

//! \}

#endif

#endif
